/* simulator for Abstract mACHINE cODE */

#include <stdio.h>
#define STACKSIZE 4000

int main(int argc,char *argv[])
{
  FILE *InStream;
  int iStack[STACKSIZE];
  int iTemp,iFrame,iBase=0,iTop=0;
  int iInstr,iArg,i;

  printf("\n Abstract Machine Code Simulator");

  if (argc>1)
  {
    if ((InStream=fopen(argv[1],"r"))==NULL)
    {
     perror(argv[1]);
     printf("\n");
     exit(1);
    }
    printf("Processing %s: \n\n",argv[1]);
  } else {
    printf("\nUsage: %s <Input file>\n\n",argv[0]);
    exit(1);
  }
  printf("Instr Arg: Stack...\n");
  while((i=fscanf(InStream,"%d",&iInstr))!=EOF)
  {
   if(i!=1)
   {
    printf("Non-integer found in input file - terminating\n");
    exit(1);
   }
   printf("%4i",iInstr);
   if(iInstr==1 || iInstr==14)
   {
    if((i=fscanf(InStream,"%d",&iArg))==EOF)
    {
     printf("Unxpected end of file.\n\n");
     exit(1);
    }
    else if (i!=1)
    {
     printf("Non-integer found in input file - terminating\n");
     exit(1);
    }
    printf("%4i",iArg);
   } else
    printf("%5s","");
   printf(": ");

   switch(iInstr)
   {
    case 1:
    {
     iStack[iTop++]=iArg;
     break;
    }
    case 2:
    {
     int iFrst=iStack[--iTop],iScnd=iStack[--iTop];
     iStack[iTop++]=iFrst+iScnd;
     break;
    }
    case 3:
    {
     int iFrst=iStack[--iTop],iScnd=iStack[--iTop];
     iStack[iTop++]=iFrst*iScnd;
     break;
    }
    case 4:
    {
     int iVal=iStack[--iTop];
     iStack[iTop++]=-iVal;
     break;
    }
    case 5:
    {
     iTop++;
     break;
    }
    case 6:
    {
     iBase=iTop;
     break;
    }
    case 7:
    {
     iTop=iBase;
     break;
    }
    case 8:
    {
     iStack[iTop]=iBase;
     break;
    }
    case 9:
    {
     iBase=iStack[iTop];
     break;
    }
    case 10:
    {
     iTemp=iStack[--iTop];
     break;
    }
    case 11:
    {
     iStack[iTop++]=iTemp;
     break;
    }
    case 12:
    {
     iFrame=iBase;
     break;
    }
    case 13:
    {
     iFrame=iStack[iFrame];
     break;
    }
    case 14:
    {
     iStack[iTop++]=iStack[iFrame+iArg];
     break;
    }
    default:
    {
     printf("Invalid instruction %i\n\n",iInstr);
     exit(1);
    }
   }
   if (iTop<0)
   {
    printf("Stack underflow occured\n\n");
    exit(1);
   }
   if (iTop>=STACKSIZE)
   {
    printf("Stack overflow occured\n\n");
    exit(1);
   }
   for(i=0;i<iTop;i++)
    printf("%4i",iStack[i]);
   printf("\n");
  }

printf("\nResult: %i\n\n",iStack[0]);
fclose(InStream);

exit(0);

}
